#include "ImageCorrelation.h"

#include "itkComplexToRealImageFilter.h"
#include "itkRescaleIntensityImageFilter.h"
#include "itkImageDuplicator.h"
#include "itkConstantPadImageFilter.h"
#include "itkMultiplyImageFilter.h"
#include "itkDivideImageFilter.h"
#include "itkDivideByConstantImageFilter.h"
#include "itkSqrtImageFilter.h"
#include "itkMinimumMaximumImageCalculator.h"
#include "itkComplexToRealImageFilter.h"
#include "itkComplexToImaginaryImageFilter.h"

#include <limits>
#include <algorithm>

#include <math.h>

#include "util.h"


typedef float RealFilterOutputType;
typedef itk::ComplexToRealImageFilter
          <ComplexImageType, ScalarImageType> RealFilterType;
typedef itk::RescaleIntensityImageFilter
          <ScalarImageType, ScalarImageType> RescaleFilterType;
typedef itk::ConstantPadImageFilter
          <ScalarImageType, ScalarImageType> PadFilterType;
typedef itk::MultiplyImageFilter
          <ComplexImageType, ComplexImageType, ComplexImageType> 
                         MultiplyComplexImageFilter;
typedef itk::DivideImageFilter
          <ComplexImageType, ComplexImageType, ComplexImageType> 
                         DivideComplexImageFilter;
typedef itk::SqrtImageFilter
          <ComplexImageType, ComplexImageType> SqrtComplexImageFilter;
typedef itk::MinimumMaximumImageCalculator<ScalarImageType> MinMaxCalculatorType;

typedef itk::ComplexToRealImageFilter
          <ComplexImageType, ScalarImageType>  ComplexImageRealFilterType;
typedef itk::ComplexToImaginaryImageFilter
          <ComplexImageType, ScalarImageType> ComplexImageImaginaryFilterType;
typedef itk::DivideByConstantImageFilter
          < ScalarImageType, float, ScalarImageType > DivideFilter;

//#define PRINT_OUTPUT 1

//
//  compare -- compare one region to another and calculate the max
//  correlation
//
ImageCorrelation::ImageCorrelation
        ( const ScalarImageType::Pointer& img1, 
          const ScalarImageType::Pointer& img2,
          FillType fillType ) 
{
   //cout << "calling compare" << endl ;

   //
   //  make copies of the images
   //
   ImageDuplicator::Pointer img_dup = ImageDuplicator::New(); 
   img_dup->SetInputImage(img1);
   img_dup->Update();
   p_image1 = img_dup->GetOutput();

   ImageDuplicator::Pointer this_dup = ImageDuplicator::New(); 
   this_dup->SetInputImage(img2);
   this_dup->Update();
   p_image2 = this_dup->GetOutput();

   //
   //  pad the images
   //
   pad( fillType );
   //char testfilename[100];
   //sprintf(testfilename, "regions/region%05d_padtest.mha", R.label());
   //save<ScalarImageType>(pad_img, testfilename);
   //cout << "image size: " << pad_img->GetLargestPossibleFFTImage() << endl;
   
   //
   //  get the scalar norms
   //
   float scalar_norm = norm( p_image1 ) * norm( p_image2 );
   //cout << "scalar_norm: " << scalar_norm << endl;
   
   //
   //  calculate the FT for each image
   //
   FFTFilterType::Pointer fftinput_F1 = FFTFilterType::New();
   fftinput_F1->SetInput(p_image1);
   fftinput_F1->Update();
   //printFFT( fftinput_F1->GetOutput(), 1 );
   //printFFTPhase( fftinput_F1->GetOutput(), R->label() );

   FFTFilterType::Pointer fftinput_F2 = FFTFilterType::New();
   fftinput_F2->SetInput(p_image2);
   fftinput_F2->Update();
   //char s[100];
   //sprintf(s, "_pad%05ld", 1 );
   //printFFTPhase( fftinput_F2->GetOutput(), label(), s );


   //
   //  calculate the complex conguate of F1 and F2
   //
   ComplexImageType::Pointer F1 = fftinput_F1->GetOutput();
   ComplexImageType::Pointer ccF1 =
                      calcComplexConjugate(fftinput_F1->GetOutput());
   ComplexImageType::Pointer F2 = fftinput_F2->GetOutput();
   ComplexImageType::Pointer ccF2 = 
                      calcComplexConjugate(fftinput_F2->GetOutput());

   //
   //  calculate the correlation
   //
   ComplexImageType::Pointer correlation = calcMult(ccF1, F2);
   //char f[100];
   //sprintf(f, "corr%06ld", R->label() );
   //printFFT( correlation, label(), string(f) );


   //
   //  calculate the inv transform of the correlation
   //
   invFFTFilterType::Pointer fftoutput = invFFTFilterType::New();
   fftoutput->SetInput(correlation); 
   fftoutput->Update();
#ifdef PRINT_OUTPUT
   char filename[100];
   sprintf(filename, "regions/region%05ld_corr%05ld.png", p_label, R->label());
   save_as_png<ScalarImageType>(fftoutput->GetOutput(), filename, true );
#endif
      

   //ScalarImageType::Pointer max_correlation_image = fftoutput->GetOutput();
   p_unnorm_corr_image = fftoutput->GetOutput();

   //
   // normalize the output
   //
   DivideFilter::Pointer divFilter = DivideFilter::New();
   divFilter->SetInput( fftoutput->GetOutput() );
   divFilter->SetConstant( scalar_norm );
   divFilter->Update();
   p_norm_corr_image = divFilter->GetOutput() ;

}

float ImageCorrelation::getUnnormalizedMaxCorrelation()
{
   MinMaxCalculatorType::Pointer calc = MinMaxCalculatorType::New();
   calc->SetImage( p_unnorm_corr_image ); 
   calc->Compute();
   return calc->GetMaximum();

}

float ImageCorrelation::getNormalizedMaxCorrelation()
{
   MinMaxCalculatorType::Pointer calc = MinMaxCalculatorType::New();
   calc->SetImage( p_norm_corr_image );
   calc->Compute();
   return calc->GetMaximum();
}



//
//  pad --  pad the images so they are square and factors of 2
//
void ImageCorrelation::pad( FillType fill_type )
{
   ScalarImageType::RegionType::SizeType size1 =
      p_image1->GetLargestPossibleRegion().GetSize();
   ScalarImageType::RegionType::SizeType size2 =
      p_image2->GetLargestPossibleRegion().GetSize();

   //
   //  how much do we pad the images?
   //
   ScalarImageType::SizeType size;
   size[0] = max( size1[0], size2[0] );
   size[1] = max( size1[1], size2[1] );

   // use the max of both for the size - make it square
   int size_test = max(size[0], size[1]);
   size[0] = size_test;  size[1] = size_test;

   //
   //  make xdist and ydist a factor of 2
   //
   //cout << "start size: " << size << endl;
   for(unsigned int i = 0; i < 2; i++)
   {
      int n = 0;
      while(size[i] >>= 1)
         n++;
      size[i] = 1 << (n+1);
   }
   //cout << "pad size: " << size << endl;
  
   //
   //  pad the images
   //

   // padding for "img"
   //cout << "image a: " << img_x_size << ", " << img_y_size << endl;
   int padx = ( size[0] - size1[0] ) / 2;
   int pady =  ( size[1] - size1[1] )/2 ;
   //cout << "padx, pady: " << padx << ", " << pady << endl ;
   unsigned long upperfactors[2] = {padx, pady}; 
   if((size[0]-size1[0])%2 != 0) padx+=1;
   if((size[1]-size1[1])%2 != 0) pady+=1;
   //cout << "padx, pady: " << padx << ", " << pady << endl ;
   unsigned long lowerfactors[2] = {padx, pady}; 

   PadFilterType::Pointer padFilter = PadFilterType::New();
   padFilter->SetInput( p_image1 );
   //unsigned long upperfactors[2] = { 10, 20};
   //unsigned long lowerfactors[2] = { 10, 20};
   if( fill_type == BLACK )
      padFilter->SetConstant(0);
   else if (fill_type == WHITE )
      padFilter->SetConstant( 255 );
   padFilter->SetPadLowerBound(lowerfactors);
   padFilter->SetPadUpperBound(upperfactors);
   padFilter->UpdateLargestPossibleRegion();
   padFilter->Update();

   p_image1 = padFilter->GetOutput();

#ifdef PRINT_OUTPUT
   char filename[100];
   //sprintf(filename, "regions/slice%02i_region%05ld_pad%05ld.mha", 
   //      p_slice_index+1, img_label, this_image_label);
   sprintf(filename, "image1_pad_%i.mha", fill_type);
   save< ScalarImageType > (p_image1, filename);
   //save_as_png<ScalarImageType>(img, filename);
#endif


   //cout << "image a: " <<  p_boundingBox.sizex() << ", " <<
   //   p_boundingBox.sizey() << endl;
   padx = (size[0] - size2[0]) / 2;
   pady =  (size[1] - size2[1]) / 2;
   //cout << "padx, pady: " << padx << ", " << pady << endl ;
   unsigned long upperfactors_B[2] = {padx, pady}; 
   if((size[0] - size2[0])%2 != 0) padx+=1;
   if((size[1] - size2[1])%2 != 0) pady+=1;
   unsigned long lowerfactors_B[2] =  {padx, pady};

   PadFilterType::Pointer padFilter_B = PadFilterType::New();
   padFilter_B->SetInput( p_image2 );
   if( (fill_type == BLACK) || (fill_type == FLOOD)  )
      padFilter_B->SetConstant(0);
   else if( fill_type == WHITE )
      padFilter_B->SetConstant( 255 );
   padFilter_B->SetPadLowerBound(lowerfactors_B);
   padFilter_B->SetPadUpperBound(upperfactors_B);
   padFilter_B->UpdateLargestPossibleRegion();
   padFilter_B->Update();

   p_image2 = padFilter_B->GetOutput();

#ifdef PRINT_OUTPUT
   //sprintf(filename, "regions/slice%02i_region%05ld_pad%05ld.mha", 
   //                  p_slice_index, this_image_label, img_label);
   sprintf(filename, "image2_pad_%i.mha", fill_type);
   save<ScalarImageType>(p_image2, filename);
   //save_as_png<ScalarImageType>(padFilter_B->GetOutput(), filename, false);
#endif

}



//
//  calcComplexConjugate -- calculate the complex conjugate of a
//  ComplexImage
//
ComplexImageType::Pointer ImageCorrelation::calcComplexConjugate
                   (const ComplexImageType::Pointer& I) const
{
   ComplexImageDuplicator::Pointer img_dup =
      ComplexImageDuplicator::New();
   img_dup->SetInputImage(I);
   img_dup->Update();
   ComplexImageType::Pointer new_img = img_dup->GetOutput();

   ComplexImageIterator iter(new_img, 
          new_img->GetLargestPossibleRegion());
   for(iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
   {
      iter.Set(std::complex<float>(iter.Get().real(), -iter.Get().imag()));
   }
   new_img->Update();

   return new_img;
}


//
//  calcSqrt -- calculate the square root of an image of type complex
//
ComplexImageType::Pointer ImageCorrelation::calcSqrt
               (const ComplexImageType::Pointer& I) const
{
   ComplexImageDuplicator::Pointer img_dup =
      ComplexImageDuplicator::New();
   img_dup->SetInputImage(I);
   img_dup->Update();
   ComplexImageType::Pointer new_img = img_dup->GetOutput();

   ComplexImageIterator iter(new_img, new_img->GetLargestPossibleRegion());
   for(iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
   {
      iter.Set(sqrt(iter.Get()));
   }
   new_img->Update();

   return new_img;
}

ComplexImageType::Pointer ImageCorrelation::calcDiv
               (const ComplexImageType::Pointer& top,
                const float& bottom) const
{
   ComplexImageDuplicator::Pointer img_dup =
      ComplexImageDuplicator::New();
   img_dup->SetInputImage(top);
   img_dup->Update();
   ComplexImageType::Pointer new_img = img_dup->GetOutput();

   ComplexImageIterator iter(new_img, new_img->GetLargestPossibleRegion());
   ComplexImageIterator iter_top(top, top->GetLargestPossibleRegion());
   iter_top.GoToBegin();
   for(iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
   {
      iter.Set(iter_top.Get() / bottom );

      ++iter_top;
   }
   new_img->Update();

   return new_img;

}


//
//  calcDiv -- calculate the division between two complex images
//
ComplexImageType::Pointer ImageCorrelation::calcDiv
               (const ComplexImageType::Pointer& top,
                const ComplexImageType::Pointer& bottom) const
{
   ComplexImageDuplicator::Pointer img_dup =
      ComplexImageDuplicator::New();
   img_dup->SetInputImage(top);
   img_dup->Update();
   ComplexImageType::Pointer new_img = img_dup->GetOutput();

   ComplexImageIterator iter(new_img, new_img->GetLargestPossibleRegion());
   ComplexImageIterator iter_top(top, top->GetLargestPossibleRegion());
   ComplexImageIterator iter_bottom(bottom, 
                             bottom->GetLargestPossibleRegion());
   iter_top.GoToBegin();
   iter_bottom.GoToBegin();
   for(iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
   {
      iter.Set(iter_top.Get() / iter_bottom.Get() );

      ++iter_top;
      ++iter_bottom;
   }
   new_img->Update();

   return new_img;

}


//
//  calcDiv -- calculate the division between a complex and scalar image
//
ComplexImageType::Pointer ImageCorrelation::calcDiv
               (const ComplexImageType::Pointer& top,
                const ScalarImageType::Pointer& bottom) const
{
   ComplexImageType::Pointer new_img = copy<ComplexImageType>(top);

   ComplexImageIterator iter(new_img, new_img->GetLargestPossibleRegion());
   ScalarImageIterator iter_bottom(bottom, 
                             bottom->GetLargestPossibleRegion());
   iter_bottom.GoToBegin();
   for(iter.GoToBegin(); !iter.IsAtEnd(); ++iter)
   {
      //iter.Set(std::complex<float>(iter.Get().real() / iter_bottom.Get(), 
      //                             iter.Get().imag() / iter_bottom.Get() ));
      iter.Set( iter.Get() / iter_bottom.Get() );

      ++iter_bottom;
   }
   new_img->Update();

   return new_img;
}

//
//  calcMult -- multiply two complex images, element-wise
//
ComplexImageType::Pointer ImageCorrelation::calcMult
               (const ComplexImageType::Pointer& I1,
                const ComplexImageType::Pointer& I2) const
{
   ComplexImageType::Pointer new_img = copy<ComplexImageType>(I1);

   ComplexImageIterator iter1(new_img, new_img->GetLargestPossibleRegion());
   ComplexImageIterator iter2(I2, I2->GetLargestPossibleRegion());
   iter2.GoToBegin();
   for(iter1.GoToBegin(); !iter1.IsAtEnd(); ++iter1)
   {
      iter1.Set( iter1.Get() * iter2.Get() );
      /*
      float a, bi, c, di;
      a = iter1.Get().real();
      bi = iter1.Get().imag();
      c = iter2.Get().real();
      di = iter2.Get().imag();
      iter1.Set(std::complex<float>((a*c)-(bi*di), (a*di)+(bi*c)));
      */

      ++iter2;
   }
   new_img->Update();

   return new_img;
}

void ImageCorrelation::normalize(ComplexImageType::Pointer& img) const
{
   ComplexImageIterator iter( img, img->GetLargestPossibleRegion() );
   std::complex< float > sum = 0;
   for( iter.GoToBegin(); !iter.IsAtEnd(); ++iter )
   {
      sum += iter.Get()*iter.Get();
   }

   sum = sqrt( sum );

   for( iter.GoToBegin(); !iter.IsAtEnd(); ++iter )
   {
      iter.Set( iter.Get() / sum );
   }

   img->Update();

}

void ImageCorrelation::normalize(ScalarImageType::Pointer& img) const
{
   ScalarImageIterator iter( img, img->GetLargestPossibleRegion() );
   float sum = 0;
   for( iter.GoToBegin(); !iter.IsAtEnd(); ++iter )
   {
      sum += iter.Get()*iter.Get();
   }

   sum = sqrt( sum );

   for( iter.GoToBegin(); !iter.IsAtEnd(); ++iter )
   {
      iter.Set( iter.Get() / sum );
   }

   img->Update();

}

float ImageCorrelation::norm( const ScalarImageType::Pointer& img ) const
{
   ScalarImageIterator iter( img, img->GetLargestPossibleRegion() );
   float sum = 0;
   for( iter.GoToBegin(); !iter.IsAtEnd(); ++iter )
   {
      sum += iter.Get()*iter.Get();
   }

   return sqrt( sum );
}


//
//  printComplexImageReal -- write the real part of the complex image to
//  a file
//
void ImageCorrelation::printComplexImageReal
         (const ComplexImageType::Pointer I, const long int& label) const
{
   ComplexImageRealFilterType::Pointer realFilter = 
                ComplexImageRealFilterType::New();
   realFilter->SetInput(I);

   char filename[100];
   sprintf(filename, "regions/region%05ld_FFTreal.png", label);
#ifdef PRINT_OUTPUT
   save_as_png<ScalarImageType>(realFilter->GetOutput(), filename);
#endif
}

//
//  printComplexImageImaginary -- write the imaginary part of the complex image to
//  a file
//
void ImageCorrelation::printComplexImageImaginary
         (const ComplexImageType::Pointer I, const long int& label) const
{
   ComplexImageImaginaryFilterType::Pointer realFilter = 
                ComplexImageImaginaryFilterType::New();
   realFilter->SetInput(I);

   char filename[100];
   sprintf(filename, "regions/region%05ld_FFTimag.png", label);
#ifdef PRINT_OUTPUT
   save_as_png<ScalarImageType>(realFilter->GetOutput(), filename);
#endif
}



void ImageCorrelation::printFFTPhase
    ( ComplexImageType::Pointer I, const long int& label, string s ) const
{
   ComplexImageIterator I_iter( I, I->GetLargestPossibleRegion() );
   I_iter.GoToBegin();
   while( !I_iter.IsAtEnd() )
   {
      I_iter.Set( I_iter.Get() / abs( I_iter.Get() ) );
      ++I_iter;
   }
   //cout << endl;

   invFFTFilterType::Pointer invfft = invFFTFilterType::New();
   invfft->SetInput( I ); 
   invfft->Update();
   char filename[ 100 ];
   sprintf( filename, "regions/phasefft_region%06ld%s.mha", label, s.c_str() );
#ifdef PRINT_OUTPUT
   save< ScalarImageType >( invfft->GetOutput(), filename );
#endif


}


void ImageCorrelation::printFFT( const ComplexImageType::Pointer& I,
                       const long int& label,
                       string s ) const
{
   ScalarImageType::Pointer image = ScalarImageType::New();
   image->SetRegions(I->GetLargestPossibleRegion().GetSize());
   image->SetOrigin(I->GetOrigin());
   image->SetSpacing(I->GetSpacing());
   image->Allocate();
   image->FillBuffer(float(0));

   ScalarImageIterator image_iter( image,
         image->GetLargestPossibleRegion() );
   ComplexImageIterator I_iter( I, I->GetLargestPossibleRegion() );
   image_iter.GoToBegin();
   I_iter.GoToBegin();
   while( !I_iter.IsAtEnd() )
   {
      //cout << I_iter.Get() << " " ;
      image_iter.Set( log( 1 + abs( I_iter.Get() )) );
      ++image_iter;
      ++I_iter;
   }
   //cout << endl;

   char filename[ 100 ];
   sprintf( filename, "regions/fft_region%06ld%s.mha", label, s.c_str() );
#ifdef PRINT_OUTPUT
   save< ScalarImageType >( image, filename );
   //save_as_png< ScalarImageType >( image, filename );
#endif


}


//
//  calcDenominator  --  shortcut for calculating F1*xF1xF2*xF2
//
ScalarImageType::Pointer ImageCorrelation::calcDenominator
         (const ComplexImageType::Pointer& F1,
          const ComplexImageType::Pointer& ccF1,
          const ComplexImageType::Pointer& F2,
          const ComplexImageType::Pointer& ccF2) const
{
   // make new "real" image
   ScalarImageType::Pointer denominator = ScalarImageType::New();
   denominator->SetRegions(F1->GetLargestPossibleRegion().GetSize());
   denominator->SetOrigin(F1->GetOrigin());
   denominator->SetSpacing(F1->GetSpacing());
   denominator->Allocate();
   denominator->FillBuffer(float(0));

   ScalarImageIterator denom_iter(denominator, 
         denominator->GetLargestPossibleRegion());
   ComplexImageIterator F1_iter(F1, F1->GetLargestPossibleRegion());
   ComplexImageIterator F2_iter(F2, F2->GetLargestPossibleRegion());
   ComplexImageIterator ccF1_iter(ccF1, ccF1->GetLargestPossibleRegion());
   ComplexImageIterator ccF2_iter(ccF2, ccF2->GetLargestPossibleRegion());
   denom_iter.GoToBegin(); F1_iter.GoToBegin(); F2_iter.GoToBegin();
   while(!denom_iter.IsAtEnd())
   {
      float a2 = pow(F1_iter.Get().real(), 2);
      float b2 = pow(F1_iter.Get().imag(), 2);
      float c2 = pow(F2_iter.Get().real(), 2);
      float d2 = pow(F2_iter.Get().imag(), 2);

      denom_iter.Set( sqrt( (a2+b2) * (c2+d2) ) + 1.0e-12);
      //denom_iter.Set( ((a2+b2)+(c2+d2))/2.0 + 1.0e-12 );
      //denom_iter.Set( 1.0 );
         
      ++denom_iter;
      ++F1_iter;
      ++F2_iter;
   }
   denominator->Update();
   
   return denominator;
}




//
//  calcNormalizedCorrelation -- actualy, we calcualte both, normalized
//  and unnormalized
//
void ImageCorrelation::calcNormalizedCorrelation
        ( const ComplexImageType::Pointer& F1,
          const ComplexImageType::Pointer& ccF1,
          const ComplexImageType::Pointer& F2,
          const ComplexImageType::Pointer& ccF2,
          ComplexImageType::Pointer& corr_image,
          ComplexImageType::Pointer& norm_corr_image,
          const float& scalar_norm ) const
{

   //
   //  calculate F1* x F2
   //
   ComplexImageType::Pointer ccF1_x_F2 = calcMult(ccF1, F2);

   //printComplexImageReal(ccF1_x_F2, label);
   //printComplexImageImaginary(ccF1_x_F2, label);

   corr_image = copy<ComplexImageType>(ccF1_x_F2);

   //
   //  calculate the normalization factor:  F1* x F1 x F2* x F2
   //
//   ScalarImageType::Pointer denom = calcDenominator(F1, ccF1, F2, ccF2);
//   ComplexImageType::Pointer div_img = calcDiv(ccF1_x_F2, denom);
   //ComplexImageType::Pointer multF1 = calcMult(ccF1,F1);
   //ComplexImageType::Pointer multF2 = calcMult(ccF2,F2);
   //ComplexImageType::Pointer denom = calcMult(multF1, multF2);
   //ComplexImageType::Pointer div_img = calcDiv(ccF1_x_F2, denom);
   ComplexImageType::Pointer div_img = calcDiv(ccF1_x_F2, scalar_norm);

   norm_corr_image = copy<ComplexImageType>(div_img);

}



