#include "Region.h"
#include "ImageCorrelation.h"

#include <vector>

Region::Region( const Region& R )
{
   p_label = R.p_label;
   p_sliceNum = R.p_sliceNum;
   p_minIndex = R.p_minIndex;
   p_maxIndex = R.p_maxIndex;

   p_centerIndex = R.p_centerIndex;
}

Region::Region( LabelType l, LabelImageType::IndexType i1, 
                             LabelImageType::IndexType i2,
                             int sliceNum )
{
   p_label = l;
   p_sliceNum = sliceNum;
   p_minIndex = i1;
   p_maxIndex = i2;

   calcRegionCenter(); // I think I should do this every time.
}

//
//  update the bounding box
//
void Region::updateIndex( LabelImageType::IndexType index )
{
   if( index[0] < p_minIndex[0] )
      p_minIndex[0] = index[0];
   if( index[1] < p_minIndex[1] )
      p_minIndex[1] = index[1];

   if( index[0] > p_maxIndex[0] )
      p_maxIndex[0] = index[0];
   if( index[1] > p_maxIndex[1] )
      p_maxIndex[1] = index[1];

   calcRegionCenter(); // I think I should do this every time.
}

//
// print the region info
//
void Region::print()
{
   cout << p_label << " " << p_minIndex << " - " << p_maxIndex << endl ;
}

ScalarImageType::SizeType Region::getScalarSize()
{
   ScalarImageType::SizeType size;
   size[0] = p_maxIndex[0] - p_minIndex[0]+1;
   size[1] = p_maxIndex[1] - p_minIndex[1]+1;

   return size;
}

LabelImageType::SizeType Region::getLabelSize()
{
   LabelImageType::SizeType size;
   size[0] = p_maxIndex[0] - p_minIndex[0];
   size[1] = p_maxIndex[1] - p_minIndex[1];

   return size;
}

ScalarImageType::IndexType Region::getScalarStart()
{
   ScalarImageType::IndexType start;
   start[0] = p_minIndex[0];
   start[1] = p_minIndex[1];

   return start;
}

LabelImageType::IndexType Region::getLabelStart()
{
   LabelImageType::IndexType start;
   start[0] = p_minIndex[0];
   start[1] = p_minIndex[1];

   return start;
}

void Region::image( ScalarImageType::Pointer image )
{
   ImageDuplicator::Pointer dup = ImageDuplicator::New();
   dup->SetInputImage( image );
   dup->Update();
   p_regionImage = dup->GetOutput();

   calcRegionCenter();
}

void Region::calcRegionCenter()
{
   p_centerIndex[0] = p_minIndex[0] + 
      ( ( p_maxIndex[0] - p_minIndex[0] ) / 2 );
   p_centerIndex[1] = p_minIndex[1] + 
      ( ( p_maxIndex[1] - p_minIndex[1] ) / 2 );
}

Correspondence Region::compare( Region& R, float dist_thresh )
{
   // compute distance
   float dist = sqrt( powf(p_centerIndex[0] - R.p_centerIndex[0], 2) +
                      powf(p_centerIndex[1] - R.p_centerIndex[1], 2)  ); 
   //cout << "indices: " << p_centerIndex << endl ;

   // compute correlation if the region is within a reasonable distance
   float maxCorr = -1 ;
   float normMaxCorr = -1 ;
   if( dist < dist_thresh )
   {
      ImageCorrelation IC( p_regionImage, R.image() );
      maxCorr = IC.getUnnormalizedMaxCorrelation() ;
      normMaxCorr = IC.getNormalizedMaxCorrelation() ;
   }


   // difference in sizes?
   int sizeA = getScalarSize()[0] * getScalarSize()[1];
   int sizeB = R.getScalarSize()[0] * R.getScalarSize()[1];
   int size = abs( sizeA - sizeB );
   //cout << "sizes: " << R.getScalarSize() << endl;
   //char fn[100]; sprintf(fn, "regions/%i_%i.mha", R.sliceNum(), R.label());
   //save< ScalarImageType >( R.image(), fn ); 

   Correspondence C( p_sliceNum, p_label, 
                     R.sliceNum(), R.label(),
                     p_centerIndex[0], p_centerIndex[1],
                     R.p_centerIndex[0], R.p_centerIndex[1],
                     maxCorr, normMaxCorr, dist, size );
   //cout << "corr: " ;
   //C.printKey();
   if( maxCorr > 0 ) C.print();
   //cout << endl;

   return C;
  
   /*
   // get the max size of each region
   ScalarImageType::SizeType max_size = getScalarSize();
   ScalarImageType::SizeType size = R.getScalarSize();
   if( max_size[0] < size[0] )
      max_size[0] = size[0];
   if( max_size[1] < size[1] )
      max_size[1] = size[1];

   cout << "size of A: " <<
      p_regionImage->GetLargestPossibleRegion().GetSize() << endl ;
   cout << "size of B: " << R.getScalarSize() << endl ;
   cout << "new size: " << max_size << endl ;

   //
   // pad the images to this size
   //
   ScalarImageType::Pointer regionA;
   ScalarImageType::Pointer regionB;
   if( max_size != getScalarSize() )
   {
      regionA = pad( max_size );
   }
   else 
      regionA = p_regionImage;
   if( max_size != R.getScalarSize() )
   {
      regionB = R.pad( max_size );
   }
   else
      regionB = R.image();
   */
    
}

/*
ScalarImageType::Pointer Region::pad( ScalarImageType::SizeType newSize )
{
   ScalarImageType::SizeType padding;
   padding[0] = (newSize[0] - p_maxIndex[0] - p_minIndex[0]) / 2;
   padding[1] = (newSize[1] - p_maxIndex[1] - p_minIndex[1]) / 2;

   PadImageFilter::Pointer pad = PadImageFilter::New();
   pad->SetConstant( 0.0 );
   pad->SetPadLowerBound( padding );
   pad->SetPadUpperBound( padding );
   pad->SetInput( p_regionImage );
   pad->UpdateLargestPossibleRegion();
   pad->Update();

   return pad->GetOutput();
}
*/

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

//
//  RegionList Constructor
//
RegionList::RegionList( LabelImageType::Pointer label_image,
                        ScalarImageType::Pointer raw_image,
                        int slice_num, 
                        const vector< LabelType >& skip_these_labels )
{
   //
   //  loop over the image and get all the labels and region locations
   //
   LabelImageIterator liter( label_image, 
         label_image->GetLargestPossibleRegion() );
   for( liter.GoToBegin(); !liter.IsAtEnd(); ++liter )
   {
      LabelType l = liter.Value(); 
      //cout << l << endl ;
      //if( l > 0 )
      // don't make regions out of the labels found in this list (which
      // is probably just 1 or 2 labels)
      if( find( skip_these_labels.begin(), skip_these_labels.end(), l ) == skip_these_labels.end() )
      {
         if( p_regionList.find( l ) == p_regionList.end() )
         {
            Region R( l, liter.GetIndex(), liter.GetIndex(), slice_num );
            p_regionList[ l ] = R;
         }
         else
         {
            // update the bounding box
            //Region &R = p_regionList[ l ];
            p_regionList[l].updateIndex( liter.GetIndex() );
         }
      }
   }
   cout << p_regionList.size() << " labels found." << endl ;

   //
   //  make the regions
   //
   RegionMapType::iterator regionIter;
   for( regionIter = p_regionList.begin();
         regionIter != p_regionList.end();
         regionIter++ )
   {
      //Region &R = regionIter->second;
      //R.print();

      // set up the region information
      ScalarImageType::SizeType size = regionIter->second.getScalarSize();
      ScalarImageType::IndexType start = regionIter->second.getScalarStart();
      ScalarImageType::RegionType region;
      region.SetSize( size );
      region.SetIndex( start );

      //  get rid of small regions - not worth caring about
      //if( size[0] > 1 )
      //{
         //cout << "extracting region " << region << endl ;

         // extract the image
         /*
         ExtractImageFilter::Pointer regionFilter =
         ExtractImageFilter::New();
         regionFilter->SetInput( raw_image );
         regionFilter->SetExtractionRegion( region );
         regionFilter->Update();
         ScalarImageType::Pointer regionImage =
         regionFilter->GetOutput();
          */

         ScalarImageType::Pointer regionImage = 
            new2d< ScalarImageType >( size[0], size[1] ); 
         RawImageIterator regionImageIter( regionImage,
               regionImage->GetLargestPossibleRegion() );
         LabelImageIterator labelImageIter( label_image, region );
         regionImageIter.GoToBegin();
         for( labelImageIter.GoToBegin(); !labelImageIter.IsAtEnd(); ++labelImageIter )
         {
            if( labelImageIter.Value() == regionIter->second.label() )
            {
               regionImageIter.Set( raw_image->GetPixel( labelImageIter.GetIndex() ) );
            }
            ++regionImageIter;
         }
         regionImage->Update();
         regionIter->second.image( regionImage );

         // black out the pixels that aren't in the mask

         // write it out to see if it worked.
#ifdef DEBUG
         char regionImageName[100];
         sprintf( regionImageName, "region%03d.mha", regionIter->second.label() );
         cout << "writing " << regionImageName << endl ;
         RawImageWriter::Pointer writer = RawImageWriter::New();
         writer->SetInput( regionImage );
         writer->SetFileName( regionImageName );
         writer->Update();
#endif
      //}
      
   }
   
}

//
//  compare lists of regions
//
CorrespondenceMatrix RegionList::compare( RegionList& rl, float dist_thresh )
{
   cout << "in compare " <<p_regionList.begin()->second.sliceNum()
      << " " << rl.begin()->second.sliceNum() << endl ;

   //
   //  initialize the correspondece matrix for these two slices
   //
   CorrespondenceMatrix correspondenceMatrix(
         /*sliceA*/    p_regionList.begin()->second.sliceNum(), 
         /*sliceB*/    rl.begin()->second.sliceNum() );
   
   Correspondence junk;
   junk.printKey();  // prints the variables to stdout

   //
   // get all the labels in two data structures
   //
   vector< LabelType > allSliceALabels;
   vector< LabelType > allSliceBLabels;
   RegionMapType::iterator regionBIter;
   // get the labels from slice B
   for( regionBIter = rl.begin();
         regionBIter != rl.end();
         regionBIter++ )
   {
      allSliceBLabels.push_back( regionBIter->second.label() );
   }

   //
   //  loop over all the regions in slice A and compare them to all the
   //  regions in slice B
   //
   RegionMapType::iterator regionAIter;
   for( regionAIter = p_regionList.begin();
         regionAIter != p_regionList.end();
         regionAIter++ )
   {
      //Region& A = regionAIter->second;
      CorrespondenceList correspondenceList( regionAIter->second.label(), 
                                             regionAIter->second.sliceNum(),
                                             regionAIter->second.centerIndex()[0],
                                             regionAIter->second.centerIndex()[1] );

      // keep track of the labels for slice A
      allSliceALabels.push_back( regionAIter->second.label() );

      // calculate all the correspondences for one region in slice A
      //cout << "comparing " << A.label() << " " ;
      RegionMapType::iterator regionBIter;
      for( regionBIter = rl.begin();
            regionBIter != rl.end();
            regionBIter++ )
      {
         //Region& B = regionBIter->second;
         //cout << B.label() << "... " << endl ; 

         Correspondence C = regionAIter->second.compare( regionBIter->second, dist_thresh );
         //if( C.getNorm() > 0 ) // if we don't calculate a correlation,
         //{                     // don't add it to the data structure
            correspondenceList.add( C );
         //}
      }

      correspondenceMatrix.add( regionAIter->second.label(), correspondenceList );
   }

   correspondenceMatrix.addLabels( allSliceALabels, allSliceBLabels );

   return correspondenceMatrix;

}
